"""Implement photonic crystal geometries."""

from typing import Callable, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np


class Geometry:
    """Geometry class."""

    def __init__(
        self,
        t1: Tuple[float, float, float],
        t2: Tuple[float, float, float],
        t3: Tuple[float, float, float],
        Nx: int = 32,
        Ny: int = 32,
        Nz: int = 32,
        eps_r: Optional[np.ndarray] = None,
        mu_r: Optional[np.ndarray] = None,
    ):
        """Initialize the geometry obj."""
        self.t1 = np.array(t1).reshape(-1, 1)
        self.t2 = np.array(t2).reshape(-1, 1)
        self.t3 = np.array(t3).reshape(-1, 1)

        self.T1 = (2 * np.pi * np.cross(self.t2, self.t3, axis=0) /
                   np.dot(t1, np.cross(self.t2, self.t3, axis=0)))
        self.T2 = (2 * np.pi * np.cross(self.t3, self.t1, axis=0) /
                   np.dot(t1, np.cross(self.t2, self.t3, axis=0)))
        self.T3 = (2 * np.pi * np.cross(self.t1, self.t2, axis=0) /
                   np.dot(t1, np.cross(self.t2, self.t3, axis=0)))

        self.eps_r = eps_r or np.ones((Nx, Ny, Nz), dtype=complex)
        self.mu_r = mu_r or np.ones((Nx, Ny, Nz), dtype=complex)

        P0, Q0, R0 = np.meshgrid(
            np.linspace(-0.5, 0.5, Nx),
            np.linspace(-0.5, 0.5, Ny),
            np.linspace(-0.5, 0.5, Nz),
        )
        self.x = P0 * self.t1[0] + Q0 * self.t2[0] + R0 * self.t3[0]
        self.y = P0 * self.t1[1] + Q0 * self.t2[1] + R0 * self.t3[1]
        self.z = P0 * self.t1[2] + Q0 * self.t2[2] + R0 * self.t3[2]

        self.epsr_f: Optional[Callable] = None
        self.mur_f: Optional[Callable] = None

    def setup(self):
        """Set the material properties."""
        if self.epsr_f:
            self.epsr_f()
        if self.mur_f:
            self.mur_f()

    def plot(self):
        """Plot ep_s and mu_r."""
        raise NotImplementedError

    def set_epsr_f(self, func: Callable):
        """Set decorator to set a eps_r profile."""
        self.epsr_f = func
        return func

    def set_mur_f(self, func: Callable):
        """Set decorator to set a mu_r profile."""
        self.mur_f = func
        return func




"""Implement the solver."""

from typing import List

import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import eigh

from .brillouinzone import BrillouinZonePath as BZPath
from .geometry import Geometry
from .utils import convmat


class Solver:
    """Implement the PWE solver."""

    def __init__(self,
                 geometry: Geometry,
                 path: BZPath,
                 P: int = 1,
                 Q: int = 1,
                 R: int = 1):
        """Initialize the PWE Solver."""
        self.geo = geometry
        self.path = path
        self.P, self.Q, self.R = P, Q, R
        self.eps_rc: np.ndarray
        self.mu_rc: np.ndarray

        self.wn: List[np.ndarray] = []
        self.modes: List[np.ndarray] = []

    def run(self):
        """Run the simulation."""
        self.geo.setup()
        self.eps_rc = convmat(self.geo.eps_r, self.P, self.Q, self.R)
        self.mu_rc = convmat(self.geo.mu_r, self.P, self.Q, self.R)

        eps_rk = np.kron(np.eye(3), self.eps_rc)
        # mu_rk = np.kron(np.eye(3), self.eps_rc)
        # epsr_ki = np.kron(np.eye(3), np.linalg.inv(self.eps_rc))
        mu_rki = np.kron(np.eye(3), np.linalg.inv(self.mu_rc))

        p = np.arange(-(self.P // 2), self.P // 2 + 1)
        q = np.arange(-(self.Q // 2), self.Q // 2 + 1)
        r = np.arange(-(self.R // 2), self.R // 2 + 1)

        P0, Q0, R0 = np.meshgrid(p, q, r)
        T1, T2, T3 = self.geo.T1, self.geo.T2, self.geo.T3

        G_k = T1 * P0.flatten() + T2 * Q0.flatten() + T3 * R0.flatten()
        beta = self.path.beta_vec
        for col in range(beta.shape[1]):
            k = beta[:, col, None] - G_k

            KX = np.diag(k[0, :])
            KY = np.diag(k[1, :])
            KZ = np.diag(k[2, :])
            K_V = np.vstack((
                np.hstack((0 * KX, -KZ, KY)),
                np.hstack((KZ, 0 * KY, -KX)),
                np.hstack((-KY, KX, 0 * KZ)),
            ))
            A = K_V @ mu_rki @ K_V
            B = eps_rk

            D, V = eigh(A, B)

            self.wn.append(np.sqrt(-np.minimum(D, 0)))
            self.modes.append(V)

    def plot_bands(self):
        """Plot a bandgap diagram."""
        beta_len = self.path.beta_vec_len
        wn = np.vstack([d for d in self.wn])
        fig, ax = plt.subplots()
        ax.plot(beta_len, wn / (2 * np.pi), "k-")
        ax.set_xticklabels(self.path.symmetry_names)
        ax.set_xticks(self.path.symmetry_locations)
        ax.set_xlim(0, self.path.symmetry_locations[-1])
        ax.set_ylim(0, 2)
        ax.set_xlabel("Bloch Wave Vector $\\beta$")
        ax.set_ylabel("Frequency [$\\frac{a \\omega}{2\\pi c}$]")
        ax.grid(True)
        plt.show()



"""This module implements utilities."""
import matplotlib.pyplot as plt
import numpy as np


def convmat(A: np.ndarray, P: int = 1, Q: int = 1, R: int = 1):
    """Create a convolution matrix."""
    NH = P * Q * R
    Nx, Ny, Nz = A.shape

    p = np.arange(-(P // 2), P//2 + 1)
    q = np.arange(-(Q // 2), Q//2 + 1)
    r = np.arange(-(R // 2), R//2 + 1)

    A = np.fft.fftshift(np.fft.fftn(A) / (Nx*Ny*Nz))

    p_0 = Nx // 2
    q_0 = Ny // 2
    r_0 = Nz // 2

    C = np.zeros((NH, NH), dtype=complex)

    for r_row in range(R):
        for q_row in range(Q):
            for p_row in range(P):
                row = r_row*Q*P + q_row*P + p_row

                for r_col in range(R):
                    for q_col in range(Q):
                        for p_col in range(P):
                            col = r_col*Q*P + q_col*P + p_col

                            p_fft = p[p_row] - p[p_col]
                            q_fft = q[q_row] - q[q_col]
                            r_fft = r[r_row] - r[r_col]

                            C[row, col] = A[p_0 + p_fft, q_0 + q_fft,
                                            r_0 + r_fft]
    return C



"""This module implements classes related to the brillouinzone."""

from typing import List, Optional, Tuple

import numpy as np
from numpy.linalg import norm


class SymmetryPoint:
    """Model of a symmetry point at the brillouinzone."""

    def __init__(self, point: Tuple[float, ...], name: str):
        """Initialize a SymmetryPoint."""
        self.point = np.array(point)
        self.name = name


class BrillouinZonePath:
    """Model of a path at the brillouinzone."""

    def __init__(
        self,
        path: List[SymmetryPoint],
        t1: Tuple[float, ...],
        t2: Optional[Tuple[float, ...]] = None,
        t3: Optional[Tuple[float, ...]] = None,
        n_points: int = 5,
    ):
        """Initialize a BrillouinZonePath."""
        self.path_points = path
        self.n_points = n_points

        self.dim = len(t1)

        self._t1 = np.pad(np.array(t1), (0, 3 - len(t1)), "constant")
        self._t2 = (np.pad(np.array(t2), (0, 3 - len(t2)), "constant")
                    if t2 else np.array([0, 1, 0]))
        self._t3 = (np.pad(np.array(t3), (0, 3 - len(t1)), "constant")
                    if t3 else np.array([0, 0, 1]))

        self._T1 = (2 * np.pi * np.cross(self._t2, self._t3) /
                    np.dot(self._t1, np.cross(self._t2, self._t3)))
        self._T2 = (2 * np.pi * np.cross(self._t3, self._t1) /
                    np.dot(self._t1, np.cross(self._t2, self._t3)))
        self._T3 = (2 * np.pi * np.cross(self._t1, self._t2) /
                    np.dot(self._t1, np.cross(self._t2, self._t3)))

        self.beta_vec = None
        self.beta_vec_len = None
        self.symmetry_names = None
        self.symmetry_locations = None
        self._calculate_beta()

    def _calculate_beta(self):
        beta_path = np.stack(
            [
                np.pad(p.point, (0, 3 - len(p.point)), "constant")
                for p in self.path_points
            ],
            axis=1,
        )
        beta_path = (beta_path[0, :] * self._T1[:, None] +
                     beta_path[1, :] * self._T2[:, None] +
                     beta_path[2, :] * self._T3[:, None])

        beta_len = np.cumsum(
            norm(np.diff(beta_path, axis=1)[:self.dim], axis=0))
        beta_len = np.pad(beta_len, (1, 0), "constant")
        beta_len_interp = np.linspace(0, beta_len[-1], self.n_points)
        beta_ix = np.interp(beta_len_interp, beta_len,
                            beta_path[0, :].flatten())
        beta_iy = np.interp(beta_len_interp, beta_len,
                            beta_path[1, :].flatten())
        beta_iz = np.interp(beta_len_interp, beta_len,
                            beta_path[2, :].flatten())

        self.beta_vec = np.vstack((beta_ix, beta_iy, beta_iz))[:self.dim, :]
        self.beta_vec_len = beta_len_interp
        self.symmetry_names = [p.name for p in self.path_points]
        self.symmetry_locations = beta_len
